(*
  Copyright 2010 Gerd Stolpmann

  This file is part of Plasma, a distributed filesystem and a
  map/reduce computation framework. Unless you have a written license
  agreement with the copyright holder (Gerd Stolpmann), the following
  terms apply:

  Plasma is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  Plasma is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with Foobar.  If not, see <http://www.gnu.org/licenses/>.

*)
(* $Id: mapred_task_exec.ml 357 2011-01-25 22:48:52Z gerd $ *)

open Mapred_def
open Mapred_tasks
open Plasma_rpcapi_aux
open Printf
open Uq_engines.Operators

let dlogf = Plasma_util.dlogf

let run_stack f s =
  while not (Stack.is_empty s) do
    f (Stack.pop s)
  done


let fragment_reader c rc file =
  let esys = Plasma_client.event_system c in
  let file' =
    List.map
      (fun (filename,block,len) ->
	 let len' =
	   if len = 0L then
	     Mapred_io.file_blocks c filename
	   else
	     len in
	 (filename,block,len')
      )
      file in
  Mapred_io.read_multiple
    c rc
    ~readers:(`Async
	       (List.map
		  (fun (filename,block,len) ->
		     (fun () ->
			dlogf
			  "fragment_reader(%s,%Ld,%Ld)" filename block len;
			let rd = Mapred_io.read_file c rc filename block len in
			eps_e (`Done rd) esys
		     )
		  )
		  file'
	       ))
    ()


let delete_file c file =
  let ht = Hashtbl.create 5 in
  List.iter (fun (name,_,_) -> Hashtbl.replace ht name ()) file;
  Hashtbl.iter
    (fun name () ->
       Mapred_io.delete_file c name
    )
    ht


let delete_files_me (me : Mapred_def.mapred_env) files =
  let c = me # open_cluster (Unixqueue.create_unix_event_system()) in
  List.iter
    (fun file ->
       delete_file c file
    )
    files;
  Plasma_client.close_cluster c


let handle_out_of_memory (me : Mapred_def.mapred_env) wfiles cleanup =
  try
    run_stack (fun f -> try f() with _ -> ()) cleanup;
    delete_files_me me wfiles;
    Unix.sleep 5;   (* hmm, give OS time to free memory *)
    `Retry_later
  with
    | error ->
	let msg =
	  sprintf "Exception in handle_out_of_memory: %s%s"
	    (Netexn.to_string error)
	    (if Printexc.backtrace_status() then
	       " - backtrace: " ^ (Printexc.get_backtrace())
	     else ""
	    ) in
	`Error msg


let record_config me jc =
object
  method bigblock_size = jc#bigblock_size
  method mr_buffer_size = me#config#mr_buffer_size
  method mr_buffer_size_tight = me#config#mr_buffer_size_tight
end


let exec_map_task me jc mj (t:map_task) ti =
  let cleanup = Stack.create() in
  let wfiles = ref [] in
  try
    let esys = Unixqueue.create_unix_event_system() in
    let c = me # open_cluster esys in
    Stack.push (fun () -> Plasma_client.close_cluster c) cleanup;

    (* The output of a map task should only be written to the local node.
       Also, be careful and set replication to 1.
     *)
    Plasma_client.configure_pref_nodes c (Plasma_client.local_identities c);


    let rc = record_config me jc in
    let reader = fragment_reader c rc t.map_input in
    Stack.push (fun () -> reader # close_in()) cleanup;
    
    let create_async prefix k =
      let name = sprintf "%s%05d" prefix k in
      Mapred_io.create_file_e ~repl:1 c name
      ++ (fun () ->
	    (* Plasma_client.abort_cluster c; - impossible in async context *)
	    wfiles := [name, 0L, 0L] :: !wfiles;
	    eps_e (`Done name) esys
	 ) in
    let create_sync prefix k =
      Plasma_client.sync (create_async prefix) k in

    let writer =
      Mapred_io.write_multiple
	c rc t.map_output_prefix (Int64.of_int me#config#mr_sort_size)
	~create_sync ~create_async () in
    Stack.push (fun () -> writer # close_out()) cleanup;
    
    mj # map me jc ti reader writer;

    run_stack (fun f -> f()) cleanup;
    
    `Ok !wfiles
  with
    | Plasma_util.Out_of_shared_memory
    | Plasma_shm.Shm_max_exceeded
    | Plasma_shm.Buf_max_exceeded ->
	Netlog.logf `Warning "Out_of_shared_memory";
	handle_out_of_memory me !wfiles cleanup
    | error ->
	let msg =
	  sprintf "Exception %s%s"
	    (Netexn.to_string error)
	    (if Printexc.backtrace_status() then
	       " - backtrace: " ^ (Printexc.get_backtrace())
	     else ""
	    ) in
	run_stack (fun f -> try f() with _ -> ()) cleanup;
	`Error msg


let ba_substring ba pos len =
  let s = String.create len in
  Netsys_mem.blit_memory_to_string ba pos s 0 len;
  s
	
let exec_sort_task_generic me jc mj (t:sort_task) =
  (* very dumb impl: read everything into an array, sort it, write it.
     This works for every record reader
   *)
  let cleanup = Stack.create() in
  let wfiles = ref [] in
  try
    dlogf "Start sort %s"
      (Mapred_tasks.string_of_task_id (`Sort t));


    let esys = Unixqueue.create_unix_event_system() in
    let c = me # open_cluster esys in
    Stack.push (fun () -> Plasma_client.close_cluster c) cleanup;

    (* The output of a sort task should only be written to the local node.
       Also, be careful and set replication to 1.
     *)
    Plasma_client.configure_pref_nodes c (Plasma_client.local_identities c);

    Mapred_io.create_file ~repl:1 c t.sort_output;
    wfiles := [t.sort_output, 0L, 0L] :: !wfiles;
    
    let rc = record_config me jc in
    let reader = fragment_reader c rc t.sort_input in
    Stack.push (fun () -> reader # close_in()) cleanup;

    let sort_limit = me#config#mr_sort_size in

    let b_data = 
      Bigarray.Array1.create Bigarray.char Bigarray.c_layout sort_limit in
    let b_eof = ref 0 in
    let b_size = ref 1024 in
    let b_end = ref 0 in
    let b_lines  = ref (Array.make !b_size 0) in
    (* let b_keys   = ref (Array.make !b_size "") in *)
    let b_hashes = ref (Array.make !b_size 0) in

    let double a null new_size =
      let old_a = !a in
      a := Array.make new_size null;
      Array.blit old_a 0 !a 0 !b_size
    in

    let extract = mj # extract_key me jc in

    ( try
	while true do
	  let line = reader # input_record() in
	  let key  = extract line in
	  let hash = Hashtbl.hash key in
	  if !b_end = !b_size then (
	    let new_size = 2 * !b_size in
	    double b_lines 0 new_size;
	    (* double b_keys "" new_size; *)
	    double b_hashes 0 new_size;
	    b_size := new_size
	  );
	  let len = String.length line in
	  let old_eof = !b_eof in
	  let new_eof = !b_eof + len in
	  if new_eof > sort_limit then
	    failwith "Amount of data exceeds sort limit";
	  Netsys_mem.blit_string_to_memory line 0 b_data !b_eof len;
	  b_eof := !b_eof + len;
	  !b_lines. ( !b_end ) <- old_eof;
	  (* !b_keys.  ( !b_end ) <- key; *)
	  !b_hashes.( !b_end ) <- hash;
	  incr b_end
	done
      with
	| End_of_file -> ()
    );

    let index_array = Array.init !b_end (fun k -> k) in
    
    let get_line i =
      let i' = i+1 in
      let offs = !b_lines.(i) in
      let offs' = if i' = !b_end then !b_eof else !b_lines.(i') in
      ba_substring b_data offs (offs' - offs) 
    in

    Mapred_sort.merge_sort
      (fun i1 i2 ->
	 let h1 = Array.unsafe_get !b_hashes i1 in
	 let h2 = Array.unsafe_get !b_hashes i2 in
	 if h1 = h2 then
	   (*
	   let k1 = Array.unsafe_get !b_keys i1 in
	   let k2 = Array.unsafe_get !b_keys i2 in
	    *)
	   let l1 = get_line i1 in
	   let l2 = get_line i2 in
	   let k1 = extract l1 in
	   let k2 = extract l2 in
	   String.compare k1 k2
	 else
	   h1 - h2   (* h1, h2 are both positive *)
      )
      index_array;

    b_hashes := [| |];
    
    let writer =
      Mapred_io.write_file c rc t.sort_output in
    Stack.push (fun () -> writer # close_out()) cleanup;
    
    Array.iter
      (fun i ->
	 writer # output_record (get_line i)
      )
      index_array;
    writer # close_out();
    dlogf "Done sort %s"
      (Mapred_tasks.string_of_task_id (`Sort t));

    reader # close_in();
    if t.sort_input_del then
      delete_file c t.sort_input;

    run_stack (fun f -> f()) cleanup;
    
    `Ok [[t.sort_output, 0L, 0L]]
  with
    | Plasma_util.Out_of_shared_memory
    | Plasma_shm.Shm_max_exceeded
    | Plasma_shm.Buf_max_exceeded ->
	Netlog.logf `Warning "Out_of_shared_memory";
	handle_out_of_memory me !wfiles cleanup
    | error ->
	let msg =
	  sprintf "Exception %s%s"
	    (Netexn.to_string error)
	    (if Printexc.backtrace_status() then
	       " - backtrace: " ^ (Printexc.get_backtrace())
	     else ""
	    ) in
	run_stack (fun f -> try f() with _ -> ()) cleanup;
	`Error msg


let rec bits_free n =
  assert(n>0);
  let n' = n lsl 1 in
  if n' < 0 then
    0
  else
    1 + bits_free n'


let exec_sort_task_text me jc mj (t:sort_task) =
  (* We read the block into a prepared buffer, split THEN into lines,
     and only record the line boundaries in a bigarray.

     This method works only with text file record readers (LF as EOL),
     and the sort limit must be at most 16G.
   *)
  let cleanup = Stack.create() in
  let wfiles = ref [] in
  try
    dlogf "Start sort %s"
      (Mapred_tasks.string_of_task_id (`Sort t));

    let esys = Unixqueue.create_unix_event_system() in
    let c = me # open_cluster esys in
    Stack.push (fun () -> Plasma_client.close_cluster c) cleanup;

    (* The output of a sort task should only be written to the local node.
       Also, be careful and set replication to 1.
     *)
    Plasma_client.configure_pref_nodes c (Plasma_client.local_identities c);

    Mapred_io.create_file ~repl:1 c t.sort_output;
    wfiles := [t.sort_output, 0L, 0L] :: !wfiles;
    
    let rc = record_config me jc in
    let reader = fragment_reader c rc t.sort_input in
    Stack.push (fun () -> reader # close_in()) cleanup;

    let sort_limit = me#config#mr_sort_size in
    if Int64.of_int sort_limit > 17179869184L (* 16G *) then
      failwith "Mapred_task_exec.exec_sort_task_text: sort_limit too large";

    let shm_mng = Plasma_client.shm_manager c in
    let shm_id = shm_mng # new_id `Buf in
    Stack.push (fun () -> shm_mng # book_allocation shm_id 0) cleanup;

    shm_mng # book_allocation shm_id sort_limit;
    let b_data = Netsys_mem.alloc_memory_pages sort_limit in
    Stack.push (fun () -> Netsys_mem.memory_unmap_file b_data) cleanup;

    (* Read data: *)
    let cursor = ref 0 in
    let dest m pos len =
      if !cursor + len > sort_limit then
	failwith "Amount of data exceeds sort limit";
      Plasma_platform.blit_memory m pos b_data !cursor len;
      cursor := !cursor + len;
      eps_e (`Done ()) esys in
    Plasma_client.sync (reader # to_any_e) dest;

    let b_eof = !cursor in

    let extract = mj # extract_key me jc in
    let p = ref 0 in

    (* First count the number of lines: *)
    let b_lines = ref 0 in

    while !p < b_eof do
      let p_lf = 
	try Platform.find_lf b_data !p b_eof 
	with Not_found -> b_eof in
      incr b_lines;
      p := p_lf+1
    done;

    (* We also allocate the b_manage array in shm, because it is less
       likely then that we allocate too much memory at once (and see
       swapping).
     *)

    let shm2_id = shm_mng # new_id `Buf in

    shm_mng # book_allocation shm2_id (8 * !b_lines);
    let b_manage = 
      Bigarray.Array1.create Bigarray.int64 Bigarray.c_layout !b_lines in
    Gc.finalise (fun _ -> shm_mng # delayed_allocation shm2_id 0) b_manage;

    (* Bits 0-29: hash
       Bits 30-63: pointer to b_data
     *)

    p := 0;
    let b_end = ref 0 in

    while !p < b_eof do
      let p_lf = 
	try Platform.find_lf b_data !p b_eof 
	with Not_found -> b_eof in
      let len = p_lf - !p in
      let line = String.create len in
      Netsys_mem.blit_memory_to_string b_data !p line 0 len;
      
      let key  = extract line in
      let hash = Hashtbl.hash key in
      
      let cell =
	Int64.logor
	  (Int64.logand (Int64.of_int hash) 0x3fff_ffffL)
	  (Int64.shift_left (Int64.of_int !p) 30) in
      b_manage.{ !b_end } <- cell;
      incr b_end;
      
      p := p_lf+1
    done;

    assert(!b_end = !b_lines);
   
    (* The lower h_bits of the elements of index_array are used to store
       parts of the hash value (needs to be the MSB of the hash value)
     *)

    let h_bits =
      min 30 (if !b_end = 0 then 1 else bits_free !b_end) in

    let h_mask =
      (1 lsl h_bits) - 1 in

    let h_drop =
      30 - h_bits in

    (* The index_array cannot live in shm. It is as large as b_manage,
       so we get already some protection against memory overallocation
     *)

    let index_array = 
      Array.init !b_end
	(fun k -> 
	   let c = Bigarray.Array1.unsafe_get b_manage k in
	   let h = Int64.to_int(Int64.logand c 0x3fff_ffffL) in
	   (k lsl h_bits) lor (h lsr h_drop)
	) in

    let get_line c i =
      let i' = i+1 in
      let offs = Int64.to_int (Int64.shift_right_logical c 30) in
      let offs' = 
	if i' = !b_end then (
	  if b_data.{ b_eof-1 } = '\n' then
	    b_eof
	  else
	    b_eof + 1
	)
	else
	  let c' = Bigarray.Array1.unsafe_get b_manage i' in
	  Int64.to_int (Int64.shift_right_logical c' 30) in
      ba_substring b_data offs (offs' - offs - 1) 
    in

    Mapred_sort.merge_sort
      (fun k1 k2 ->
	 let g1 = k1 land h_mask in
	 let g2 = k2 land h_mask in
	 if g1 = g2 then (
	   let i1 = k1 lsr h_bits in
	   let i2 = k2 lsr h_bits in
	   let c1 = Bigarray.Array1.unsafe_get b_manage i1 in
	   let c2 = Bigarray.Array1.unsafe_get b_manage i2 in
	   let h1 = Int64.logand c1 0x3fff_ffffL in
	   let h2 = Int64.logand c2 0x3fff_ffffL in
	   if h1 = h2 then
	     let l1 = get_line c1 i1 in
	     let l2 = get_line c2 i2 in
	     let key1 = extract l1 in
	     let key2 = extract l2 in
	     String.compare key1 key2
	   else
	     Int64.to_int (Int64.sub h1 h2)   (* h1, h2 are both positive *)
	 )
	 else
	   g1 - g2
      )
      index_array;

    let writer =
      Mapred_io.write_file c rc t.sort_output in
    Stack.push (fun () -> writer # close_out()) cleanup;
    
    Array.iter
      (fun k ->
	 let i = k lsr h_bits in
	 let c = Bigarray.Array1.unsafe_get b_manage i in
	 writer # output_record (get_line c i)
      )
      index_array;
    writer # close_out();
    dlogf "Done sort %s"
      (Mapred_tasks.string_of_task_id (`Sort t));

    reader # close_in();
    if t.sort_input_del then
      delete_file c t.sort_input;

    run_stack (fun f -> f()) cleanup;
    
    `Ok [[t.sort_output, 0L, 0L]]
  with
    | Plasma_util.Out_of_shared_memory
    | Plasma_shm.Shm_max_exceeded
    | Plasma_shm.Buf_max_exceeded ->
	Netlog.logf `Warning "Out_of_shared_memory";
	handle_out_of_memory me !wfiles cleanup
    | error ->
	let msg =
	  sprintf "Exception %s%s"
	    (Netexn.to_string error)
	    (if Printexc.backtrace_status() then
	       " - backtrace: " ^ (Printexc.get_backtrace())
	     else ""
	    ) in
	run_stack (fun f -> try f() with _ -> ()) cleanup;
	`Error msg


type shuffle_in =
    { in_index : int;   (* which input file *)
      in_key : string;  (* current key *)
      in_hash : int;    (* hash of the key *)
    }


module Rd = struct
  type t = shuffle_in

  (* This _must_ be the same criterion as for sorting! *)
  let compare (r1:t) (r2:t) =
    if r1.in_hash = r2.in_hash then (
      let d = String.compare r1.in_key r2.in_key in
      if d = 0 then
	r1.in_index - r2.in_index
      else
	d
    )
    else
      r1.in_hash - r2.in_hash
end

module RdSet0 = Set.Make(Rd)


(* This is only faster for very small sets (maybe < 4 elements) *)
(*
module RdSet = struct
  type t = shuffle_in list

  let empty = []

  let min_elt s =
    match s with
      | [] -> raise Not_found
      | x :: _ -> x

  let remove x s =
    let rec rm s =
      match s with
	| [] -> []
	| y :: s' -> 
	    let p = Rd.compare x y in
	    if (* x > y *) p > 0 then y :: rm s' else 
	      if p = 0 then s' else s
    in
    rm s

  let add x s =
    let rec ins s =
      match s with
	| [] -> [x]
	| y :: s' ->
	    let p = Rd.compare x y in
	    if (* x > y *) p > 0 then y :: ins s' else
	      if p = 0 then x :: s' else x :: s
    in
    ins s
end
 *)

module RdSet = struct
  type t =
    | Empty
    | S1 of shuffle_in
    | S2 of shuffle_in * shuffle_in
    | S3 of shuffle_in * shuffle_in * shuffle_in
    | S4 of shuffle_in * shuffle_in * shuffle_in * shuffle_in
    | Sn of RdSet0.t

  let empty = Empty

  let min_elt =
    function
      | Empty -> raise Not_found
      | S1 x -> x
      | S2(x,_) -> x
      | S3(x,_,_) -> x
      | S4(x,_,_,_) -> x
      | Sn s -> RdSet0.min_elt s

  let remove x s =
    match s with
      | Empty -> Empty
      | S1 y0 -> if Rd.compare x y0 = 0 then Empty else s
      | S2(y0,y1) -> 
	  let p0 = Rd.compare x y0 in
	  if p0 = 0 then S1 y1 
	  else if p0 > 0 && Rd.compare x y1 = 0 then S1 y0 
	  else s
      | S3(y0,y1,y2) ->
	  let p1 = Rd.compare x y1 in
	  if p1 = 0 then S2(y0, y2) else 
	    if p1 < 0 then (
	      if Rd.compare x y0 = 0 then S2(y1,y2) else s
	    )
	    else (
	      if Rd.compare x y2 = 0 then S2(y0,y1) else s
	    )
      | S4(y0,y1,y2,y3) ->
	  let p1 = Rd.compare x y1 in
	  if p1 = 0 then S3(y0, y2, y3) else 
	    if p1 < 0 then (
	      if Rd.compare x y0 = 0 then S3(y1,y2,y3) else s
	    )
	    else (
	      let p2 = Rd.compare x y2 in
	      if p2 = 0 then S3(y0,y1,y3) else
		if p2 < 0 then s else 
		  if Rd.compare x y3 = 0 then
		    S3(y0,y1,y2)
		  else s
	    )
      | Sn u -> 
	  let u' = RdSet0.remove x u in
	  (* we do not downgrade to the tuple representation. But we need
	     to check for emptiness (for tests against empty)
	   *)
	  if u' = RdSet0.empty then
	    Empty
	  else
	    Sn u'

  let upgrade l =
    Sn (List.fold_left (fun acc x -> RdSet0.add x acc) RdSet0.empty l)

  let add x s =
    match s with
      | Empty -> S1 x
      | S1 y0 -> 
	  let p0 = Rd.compare x y0 in
	  if p0 = 0 then S1 x else 
	    if p0 < 0 then S2(x,y0) else S2(y0,x)
      | S2(y0,y1) -> 
	  let p0 = Rd.compare x y0 in
	  if p0 = 0 then S2(x, y1) else
	    if p0 < 0 then S3(x,y0,y1) else (
	      let p1 = Rd.compare x y1 in
	      if p1 = 0 then S2(y0,x) else
		if p1 < 0 then S3(y0,x,y1) else S3(y0,y1,x)
	    )
      | S3(y0,y1,y2) ->
	  let p1 = Rd.compare x y1 in
	  if p1 = 0 then S3(y0, x, y2) else
	    if p1 < 0 then (
	      let p0 = Rd.compare x y0 in
	      if p0 = 0 then S3(x,y1,y2) else
		if p0 < 0 then S4(x,y0,y1,y2) else S4(y0,x,y1,y2)
	    )
	    else (
	      let p2 = Rd.compare x y2 in
	      if p2 = 0 then S3(y0,y1,x) else
		if p2 < 0 then S4(y0,y1,x,y2) else S4(y0,y1,y2,x)
	    )
      | S4(y0,y1,y2,y3) ->
	  let up() = upgrade [x; y0; y1; y2; y3] in
	  let p1 = Rd.compare x y1 in
	  if p1 = 0 then S4(y0,x,y2,y3) else
	    if p1 < 0 then (
	      let p0 = Rd.compare x y0 in
	      if p0 = 0 then S4(x,y1,y2,y3) else
		up()
	    )
	    else (
	      let p2 = Rd.compare x y2 in
	      if p2 = 0 then S4(y0,y1,x,y3) else
		if p2 < 0 then
		  up()
		else
		  if Rd.compare x y3 = 0 then
		    S4(y0,y1,y2,x)
		  else 
		    up()
	    )
      | Sn u -> 
	  let u' = RdSet0.add x u in
	  Sn u'
end


let merging_reader input_array extract esys : Mapred_io.record_reader =
  let input_q =
    Array.map
      (fun rd ->
	 let q = Queue.create() in
	 (try rd # input_records q with End_of_file -> ());
	 q
      )
      input_array in
  let peek intr k =
    let q = input_q.(k) in
    if Queue.is_empty q then (
      intr := true;
      let rd = input_array.(k) in
      try rd # input_records q with End_of_file -> ()
    );
    Queue.peek q in
  let inprec k =
    let q = input_q.(k) in
    if Queue.is_empty q then (
      let rd = input_array.(k) in
      try rd # input_records q with End_of_file -> ()
    );
    Queue.take q in

  let input_set = ref RdSet.empty in
  let () =
    let dummy = ref false in
    for k = 0 to Array.length input_array - 1 do
      try
	let line = peek dummy k in  (* or Queue.Empty *)
	let key = extract line in
	let si =
	  { in_index = k;
	    in_key = key;
	    in_hash = Hashtbl.hash key
	  } in
	input_set := RdSet.add si !input_set
      with
	| Queue.Empty -> ()
    done in

  let pos_in = ref 0 in
  let current = Queue.create() in

  let advance() =
    assert(Queue.is_empty current);
    if !input_set <> RdSet.empty then (
      let si = RdSet.min_elt !input_set in
      let key = si.in_key in
      let k = si.in_index in
      let line = inprec k in
      (* This inprec cannot raise Queue.Empty because we have already
	 peeked the line (this is an invariant for all [si] in [input_set])
       *)
      Queue.add line current;
      ( try
	  let intr = ref false in
	  while not !intr do
	    let next_line = peek intr k in (* or Queue.Empty *)
	    let next_key = extract next_line in
	    if next_key = key then (
	      if not !intr then (
		ignore(inprec k);
		Queue.add next_line current;
	      )
	    )
	    else (
	      input_set := RdSet.remove si !input_set;
	      let si' =
		{ si with 
		    in_key = next_key; 
		    in_hash = Hashtbl.hash next_key
		} in
	      input_set := RdSet.add si' !input_set;
	      intr := true
	    )
	  done
	with
	  | Queue.Empty ->
	      input_set := RdSet.remove si !input_set
      )
    ) in

  let () = 
    advance() in

  ( object(self)
      method pos_in = !pos_in
      method input_record() =
 	if Queue.is_empty current then advance();
	try
	  let line = Queue.take current in
	  incr pos_in;
	  line
	with
	  | Queue.Empty -> raise End_of_file

      method input_records q =
	if Queue.is_empty current then advance();
	if Queue.is_empty current then raise End_of_file;
	pos_in := !pos_in + Queue.length current;
	Queue.transfer current q

      method peek_record() =
 	if Queue.is_empty current then advance();
	try
	  Queue.peek current
	with
	  | Queue.Empty -> raise End_of_file
	      
      method close_in() =
	Array.iter (fun rd -> rd # close_in()) input_array

      method abort() =
	Array.iter (fun rd -> rd # abort()) input_array

      method to_any_e dest =
	assert false
	
      method to_fd_e fd = 
	let io_dev =
	  `Polldescr(Netsys.get_fd_style fd,fd,esys) in
	let input_many_records () =
	  let l = ref [] in
	  let n = ref 0 in
	  ( try
	      while !n < 1_000_000 do
		let line = self#input_record() in
		l := line :: !l;
		n := !n + String.length line + 1;
	      done;
	    with End_of_file -> ()
	  );
	  !n, String.concat "\n" (List.rev !l) in

	dlogf
	  "Mapred_task_exec.exec_shuffle_task: reader to_fd_e";
	let stream =
	  Stream.from
	    (fun _ ->
	       let n, lines = input_many_records() in
	       dlogf
		 "Mapred_task_exec.exec_shuffle_task: reader to_fd_e n=%d" n;
	       (* Unixqueue.Debug.enable := true; *)
	       if n > 0 then (
		 let e() = 
		   Uq_io.output_string_e io_dev lines
		   ++ (fun () -> Uq_io.output_string_e io_dev "\n") in
		 Some e
	       )
	       else (
		 dlogf
		   "Mapred_task_exec.exec_shuffle_task: reader to_fd_e EOF";
		 None
	       )
	    ) in
	Uq_engines.stream_seq_engine () stream esys

      method plasma_cluster = input_array.(0) # plasma_cluster
    end
  )


let exec_shuffle_task me jc mj (t:shuffle_task) ti =
  let cleanup = Stack.create() in
  let wfiles = ref [] in
  try
    let esys = Unixqueue.create_unix_event_system() in
    let c = me # open_cluster esys in
    Stack.push (fun () -> Plasma_client.close_cluster c) cleanup;

    (* In some situations we need a cluster that can always be used
       synchronously. We can set it to [c] when we know that [c] is
       never used asynchronously (or otherwise there is no conflicting use).
     *)
    let csync =
      if t.shuffle_reduce && List.length t.shuffle_input > 1 then (
	dlogf
	  "Mapred_task_exec.exec_shuffle_task: using separate csync client";
	let c1 = me # open_cluster (Unixqueue.create_unix_event_system()) in
	Stack.push (fun () -> Plasma_client.close_cluster c1) cleanup;
	c1
      ) else
	(* if not t.shuffle_reduce: [reduce] is not called, so no danger.
	   if len t.shuffle_input = 1: No merge happens. Even if [reduce]
           is an asynchronous procedure, we know that there is no
	     synchronous line-by-line reading
	 *)
	c in

    (* Local node is always preferred *)
    Plasma_client.configure_pref_nodes c (Plasma_client.local_identities c);

    let rc = record_config me jc in

    let (pmin,pmax) = t.shuffle_partitions in
    if pmin > pmax then failwith "Shuffle: empty set of output partitions";

    (* Maybe we can shuffle with [link]: *)
    let link_possible =
       (match t.shuffle_input with         (* exactly one input *)
	  | [ [_, 0L, (*0L*) _], _, _ ] -> true
	      (* FIXME: not checking the end block is wrong. However, we
		 never get a 0 here
	       *)
	  | _ -> false
       ) && 
	 ( match t.shuffle_output with     (* exactly one output *)
	     | [ _, _, _ ] -> true
	     | _ -> false
	 ) &&
	 not t.shuffle_reduce in          (* and no reduce! *)
    
    let output_files =
      if link_possible then (
	let in_file, _, _ = List.hd t.shuffle_input in
	let in_name, _, _ = List.hd in_file in
	let out_name, _, _ = List.hd t.shuffle_output in
	Plasma_client.retry c "shuffle_task_link_possible"
	  (fun () ->
	     Plasma_client.with_trans c
               (fun trans ->
		  let inode = Plasma_client.lookup trans in_name false in
		  Plasma_client.link trans out_name inode;
	       )
	  ) ();
	wfiles := [ out_name, 0L, 0L ] :: !wfiles;
	[ [ out_name, 0L, 0L ] ]
      )
      else (
	let inputs =
	  List.map
	    (fun (file,_,_) ->
	       let reader = fragment_reader csync rc file in
	       Stack.push (fun () -> reader # close_in()) cleanup;
	       reader
	    )
	    t.shuffle_input in
	let input_array = Array.of_list inputs in
	
	let extract =
	  mj # extract_key me jc in
	let part_key =
	  mj # partition_of_key me jc in
	
	let input_files = Hashtbl.create 5 in
	List.iter2
	  (fun (file,_,_) rd ->
	     let s =
	       String.concat "+"
		 (List.map 
		    (fun (name,pos,len) -> sprintf "%s@%Ld/%Ld" name pos len)
		    file) in
	     Hashtbl.add input_files rd ("[" ^ s ^ "]")
	  )
	  t.shuffle_input
	  inputs;
	
	let output_part = 
	  Array.make (pmax+1-pmin) None in
	
	let outputs =
	  List.map
	    (fun (filename,fpmin,fpmax) ->
	       if fpmin < pmin || fpmin > fpmax || fpmax > pmax then
		 failwith "Shuffle: bad output partitions";
	       let repl = if t.shuffle_reduce then 0 else 1 in
	       Mapred_io.create_file ~repl c filename;
	       wfiles := [filename,0L,0L] :: !wfiles;
	       let writer =
	       Mapred_io.write_file c rc filename in
	       Stack.push (fun () -> writer # close_out()) cleanup;
	       for p = fpmin to fpmax do
		 output_part.(p - pmin) <- Some writer
	       done;
	       writer
    	    )
	    t.shuffle_output in
	
	let output_files =
	  List.map
	    (fun (filename,fpmin,fpmax) -> [filename, 0L, 0L])
	    t.shuffle_output in
	
	let reader =
	  if Array.length input_array > 1 then
	    merging_reader input_array extract esys  (* esys: only for to_fd_e *)
	  else
	    input_array.(0) in
	
	if t.shuffle_reduce then (
	  (** this is a merge+reduce pass *)
	  (* assert(pmin = pmax); - not true anymore *)
	  assert(List.length outputs=1);
	  let writer = List.hd outputs in
	  mj # reduce me jc ti reader writer;
	)
	else (
	  (** this is a merge+split pass *)
	  try
	    while true do
	      let line = reader # input_record() in
	      let key = extract line in
	      let p = part_key key in
	      
	      if p < pmin || p > pmax then
		failwith 
		  (sprintf "Shuffle: Partition outside allowed range \
                          (key=%s p=%d)" key p
		  );
	      
	      ( match output_part.(p - pmin) with
		  | None ->
		      failwith "Shuffle: output partition missing"
		  | Some writer ->
		      writer # output_record line
	      );
	    done
	  with End_of_file -> ()
	);
	
	List.iter (fun reader -> reader # close_in()) inputs;
	List.iter (fun writer -> writer # close_out()) outputs;
	output_files
      ) in

    if t.shuffle_input_del then (
      List.iter
	(fun (file,_,_) -> delete_file csync file)
	t.shuffle_input
    );

     run_stack (fun f -> f()) cleanup;
    
    `Ok output_files
  with
    | Plasma_util.Out_of_shared_memory 
    | Plasma_shm.Shm_max_exceeded
    | Plasma_shm.Buf_max_exceeded ->
	Netlog.logf `Warning "Out_of_shared_memory";
	handle_out_of_memory me !wfiles cleanup
    | error ->
	let msg =
	  sprintf "Exception %s%s"
	    (Netexn.to_string error)
	    (if Printexc.backtrace_status() then
	       " - backtrace: " ^ (Printexc.get_backtrace())
	     else ""
	    ) in
	run_stack (fun f -> try f() with _ -> ()) cleanup;
	`Error msg


let execute (me:Mapred_def.mapred_env) 
            (jc:Mapred_def.mapred_job_config)
            (mj:Mapred_def.mapred_job) t ti
    : task_result =
  Netlog.logf `Notice "Start task %s"
    (Mapred_tasks.string_of_task_id t);
  let r =
    match t with
      | `Map t' -> 
	  exec_map_task me jc mj t' ti
      | `Sort t' ->
	  exec_sort_task_text me jc mj t'
      | `Shuffle t' ->
	  exec_shuffle_task me jc mj t' ti
      | `Cleanup ->
	  `Ok []   (* no cleanup here. This is done in mapred_task_server *)
  in
  Netlog.logf `Notice "End task %s"
    (Mapred_tasks.string_of_task_id t);
  r
