(*
  Copyright 2010 Gerd Stolpmann

  This file is part of Plasma, a distributed filesystem and a
  map/reduce computation framework. Unless you have a written license
  agreement with the copyright holder (Gerd Stolpmann), the following
  terms apply:

  Plasma is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  Plasma is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with Foobar.  If not, see <http://www.gnu.org/licenses/>.

*)
(* $Id: nn_datanode_ctrl.ml 337 2011-01-07 21:10:13Z gerd $ *)

open Uq_engines.Operators
open Printf

let dlogf = Plasma_util.dlogf

module Rpcapi_clnt = Pfs_rpcapi_clnt
module Rpcapi_srv = Pfs_rpcapi_srv
module Rpcapi_aux = Pfs_rpcapi_aux

module Dnctrl_proxy = Rpcapi_clnt.Make'Datanode_ctrl(Rpc_proxy.ManagedClient)

type client = Rpc_proxy.ManagedClient.mclient

type ticket_verifier =
    { tv_safetrans_id : int64;
      tv_safetrans_tmo : int64;
      tv_safetrans_sec : int64;
      tv_datanodes : int list;
      tv_enabled : bool;
    }

type ctrl =
    { clients : (int, Nn_datastores.datastore * client) Hashtbl.t;
      (** maps datastore ID to client *)

      preallocated : ticket_verifier Queue.t;
      (** the queue of preallocated verifiers *)

      preallocated_size : int ref;
      (** the size to which preallocated is filled regularly *)

      preallocated_incr : bool ref;
      (** whether preallocated_size was increased *)

      disable : ticket_verifier Queue.t;
      (** the queue of verifiers to disable *)

      next_id : int64 ref;
      (** Next verifier ID *)

      datastores : Nn_datastores.datastore list ref;
      (** the current list of datastores *)

      hkeep : bool ref;
      (** whether housekeeping is in progress *)
    }

module Ctrl_var = 
  Netplex_cenv.Make_var_type(struct type t = ctrl end)

let ctrlname = "current_datanode_ctrl"

let get_ctrl() =
  Ctrl_var.get ctrlname

let get_next_id() =
  let ctrl = get_ctrl() in
  let next_id = !(ctrl.next_id) in
  ctrl.next_id := Int64.succ next_id;
  next_id


let create_client conn esys =
  let mclient_config =
    Rpc_proxy.ManagedClient.create_mclient_config
      ~programs:[ Rpcapi_clnt.Datanode_ctrl.V1._program ]
      ~msg_timeout:20.0  (* this should be quite short *)
      ~msg_timeout_is_fatal:true
      ~initial_ping:true
      () in
  let mclient =
    Rpc_proxy.ManagedClient.create_mclient
      mclient_config
      conn
      esys in
  mclient


let live_datastores() =
  (** Returns the sorted list of live datastores *)
  let ds_list =
    Nn_datastore_news.list ~wait:true () in
  List.sort
    (fun ds1 ds2 ->
       compare ds1.Nn_datastores.id ds2.Nn_datastores.id
    )
    (List.filter
       (fun ds ->
	  let identity = ds.Nn_datastores.identity in
	  ds.Nn_datastores.enabled && Nn_datastore_news.is_alive identity
       )
       ds_list
    )


let upd_datastores ctrl =
  (** Updates the [datastores] component in ctrl, and resets [preallocated]
      if there is a change.
   *)
  let dlist = live_datastores() in
  if dlist <> !(ctrl.datastores) then (
    ctrl.datastores := dlist;
    Queue.transfer ctrl.preallocated ctrl.disable
  )


let create_ticket_verifier ds_list enabled =
  { tv_safetrans_id = get_next_id();
    tv_safetrans_tmo = Int64.add (Int64.of_float (Unix.time())) 3600L; (* 1h *)
    tv_safetrans_sec = Plasma_rng.random_int64();
    tv_datanodes = List.map (fun ds -> ds.Nn_datastores.id) ds_list;
    tv_enabled = enabled;
  }


let new_ticket_verifier() =
  let ctrl = get_ctrl() in
  upd_datastores ctrl;
  let now = Unix.time() in
  let ds_list = !(ctrl.datastores) in
  let ds_id_list = List.map (fun ds -> ds.Nn_datastores.id) ds_list in
  try
    let found = ref None in
    while !found = None do
      let tv = Queue.take ctrl.preallocated in
      let min_tmo = Int64.add (Int64.of_float now) 1800L (* 30 min *) in
      if tv.tv_safetrans_tmo >= min_tmo && tv.tv_datanodes = ds_id_list then
	found := Some tv
    done;
    match !found with
      | None -> assert false
      | Some tv -> tv
  with
    | Queue.Empty ->
	if not !(ctrl.preallocated_incr) then (
	  ctrl.preallocated_size := min 65536 (2 * !(ctrl.preallocated_size));
	  ctrl.preallocated_incr := true
	);
	create_ticket_verifier ds_list false


let reset_new_datastores() =
  (** Create connections to newly added datastores, and send a reset to them *)
  let esys = (Netplex_cenv.self_cont()) # event_system in
  let ctrl = get_ctrl() in
  upd_datastores ctrl;
  let ds_list = !(ctrl.datastores) in
  List.iter
    (fun ds ->
       let id = ds.Nn_datastores.id in
       if not (Hashtbl.mem ctrl.clients id) then (
	 let conn =
	   match ds.Nn_datastores.node with
	     | None -> assert false
	     | Some addr -> Plasma_util.connector_of_sockaddr addr in
	 let client = create_client conn esys in
	 Dnctrl_proxy.V1.reset_all_safetrans'async
	   client
	   ()
	   (fun get_reply ->
	      try let () = get_reply() in ()
	      with error ->
		Netlog.logf `Err
		  "reset_all_safetrans: Exception %s"
		  (Netexn.to_string error);
		Rpc_proxy.ManagedClient.enforce_unavailability client;
		Hashtbl.remove ctrl.clients id
	   );
	 Hashtbl.add ctrl.clients id (ds,client)
       )
    )
    ds_list
  

let disconnect id =
  (** Look the client up, and terminate the connection: *)
  let ctrl = get_ctrl() in
  try
    let _, client = Hashtbl.find ctrl.clients id in
    Rpc_proxy.ManagedClient.trigger_shutdown client (fun () -> ());
    Hashtbl.remove ctrl.clients id
  with
    | Not_found -> ()


let reset() =
  (** First terminate all connections: *)
  let ctrl = get_ctrl() in
  Hashtbl.iter
    (fun id (_,client) ->
       Rpc_proxy.ManagedClient.trigger_shutdown client (fun () -> ())
    )
    ctrl.clients;
  Hashtbl.clear ctrl.clients;

  (** Get new list of datastores, create new clients, and send the reset: *)
  reset_new_datastores()


let enable_ticket_verifier_e tv =
  let esys = (Netplex_cenv.self_cont()) # event_system in
  let ctrl = get_ctrl() in
  reset_new_datastores();

  if tv.tv_enabled then
    eps_e (`Done()) esys
  else (
    Plasma_util.msync_engine
      ( List.map
	  (fun id ->
	     try
	       let _, client = Hashtbl.find ctrl.clients id in(* or Not_found *)
	       Plasma_util.rpc_engine
		 client
		 Dnctrl_proxy.V1.safetrans'async
		 (tv.tv_safetrans_id, 
		  tv.tv_safetrans_tmo, 
		  tv.tv_safetrans_sec)
	       >> (function
		     | `Done _ ->
			 dlogf
			   "Datastore %d: created safetrans %Ld secret %Ld"
			   id tv.tv_safetrans_id tv.tv_safetrans_sec;
			 `Done()
		     | `Error e ->
			 Netlog.logf `Err
			   "safetrans: Exception %s" (Netexn.to_string e);
			 `Error e
		     | `Aborted -> `Aborted
		  )
	     with
	       | Not_found ->
		   eps_e (`Error (Failure
				    (sprintf 
				       "Nn_datanode_ctrl: datanode %d is not \
                                        available anymore - cannot enable \
                                        ticket"
				       id))) esys
	  )
	  tv.tv_datanodes
      )
      esys
  )


let disable_ticket_verifier_e tv =
  (* We just add tv to the queue of verifiers to disable... *)
  let esys = (Netplex_cenv.self_cont()) # event_system in
  let ctrl = get_ctrl() in
  Queue.add tv ctrl.disable;
  eps_e (`Done ()) esys


let housekeeping() =
  (* This is called every second to
     - refill missing ticket verifiers
     - disable verifiers
   *)
  let esys = (Netplex_cenv.self_cont()) # event_system in
  let ctrl = get_ctrl() in

  let rec cancel() =
    try
      let tv = Queue.take ctrl.disable in
      let tv_elist = cancel_tv tv tv.tv_datanodes in
      tv_elist @ cancel()
    with
      | Queue.Empty ->
	  []

  and cancel_tv tv id_list =
    match id_list with
      | id :: id_list' ->
	  let e =
	    try
	      let (ds,client) = Hashtbl.find ctrl.clients id in
	      Plasma_util.rpc_engine
		client
		Dnctrl_proxy.V1.cancel_safetrans'async
		tv.tv_safetrans_id
	      >> (function
		     | `Done _ ->
			 dlogf
			   "Datastore %d: cancelled safetrans %Ld secret %Ld"
			   id tv.tv_safetrans_id tv.tv_safetrans_sec;
			 `Done()
		     | `Error e ->
			 Netlog.logf `Err
			   "cancelling safetrans: Exception %s"
			   (Netexn.to_string e);
			 `Done()   (* We ignore errors here *)
		     | `Aborted -> `Aborted
		  )
	    with
	      | Not_found ->
		  eps_e (`Done ()) esys in
	  e :: cancel_tv tv id_list' 
      | [] ->
	  []
  in

  let new_verifiers = Queue.create() in

  let rec refill ds_list =
    let n = Queue.length ctrl.preallocated + Queue.length new_verifiers in
    if n < !(ctrl.preallocated_size) then
      let tv = create_ticket_verifier ds_list true in
      Queue.add tv new_verifiers;
      let elist = refill_tv tv ds_list in
      elist @ refill ds_list
    else
      []

  and refill_tv tv ds_list =
    match ds_list with
      | ds :: ds_list' ->
	  let e =
	    try
	      let id = ds.Nn_datastores.id in
	      let (ds,client) = Hashtbl.find ctrl.clients id in
	      Plasma_util.rpc_engine
		client
		Dnctrl_proxy.V1.safetrans'async
		(tv.tv_safetrans_id, 
		 tv.tv_safetrans_tmo, 
		 tv.tv_safetrans_sec)
	      >> (function
		     | `Done _ ->
			 dlogf
			   "Datastore %d: refilled safetrans %Ld secret %Ld"
			   id tv.tv_safetrans_id tv.tv_safetrans_sec;
			 `Done()
		     | `Error e ->
			 Netlog.logf `Err
			   "refilling safetrans: Exception %s"
			   (Netexn.to_string e);
			 `Error e
		     | `Aborted -> `Aborted
		  )
	    with
	      | Not_found ->
		  eps_e (`Done ()) esys in
	  e :: refill_tv tv ds_list' 
      | [] ->
	  []
  in

  if not !(ctrl.hkeep) then (
    reset_new_datastores();
    let ds_list = !(ctrl.datastores) in
    ctrl.hkeep := true;
    let cancel_elist = cancel() in
    let refill_elist = refill ds_list in
    let _e =
      Plasma_util.msync_engine (cancel_elist @ refill_elist) esys
      ++ (fun () ->
	    ctrl.preallocated_incr := false;
	    Queue.transfer new_verifiers ctrl.preallocated;
	    ctrl.hkeep := false;
	    eps_e (`Done ()) esys
	 ) in
    ()
  )

(* FIXME: if only one allocation/deallocation goes wrong, the whole
   housekeeping round fails
 *)


let post_start_hook() =
  Ctrl_var.set ctrlname
    { clients = Hashtbl.create 7;
      preallocated = Queue.create();
      preallocated_size = ref 100;
      preallocated_incr = ref false;
      disable = Queue.create();
      next_id = ref 1L;
      datastores = ref [];
      hkeep = ref false
    };
  let _timer =
    Netplex_cenv.create_timer
      (fun _ -> 
	 (try 
	    housekeeping()
	  with
	    | error ->
		Netlog.logf `Crit
		  "Nn_datanode_ctrl: Exception during housekeeping: %s"
		  (Netexn.to_string error)
	 );
	 true
      ) 
      1.0 in
  ()



