(*
  Copyright 2010 Gerd Stolpmann

  This file is part of Plasma, a distributed filesystem and a
  map/reduce computation framework. Unless you have a written license
  agreement with the copyright holder (Gerd Stolpmann), the following
  terms apply:

  Plasma is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  Plasma is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with Foobar.  If not, see <http://www.gnu.org/licenses/>.

*)
(* $Id: nn_inodecache.ml 332 2011-01-04 01:51:32Z gerd $ *)

(** The inodecache checks whether inode metadata is still up to date.
    It requests notifications from the main nameserver.
 *)

open Pfs_rpcapi_aux
open Plasma_util.Operators

let dlogf = Plasma_util.dlogf

type value =
    [ `Deleted
    | `Exists of inodeinfo
    | `Unknown
    ]

type entry =
    { inode : int64;
      inodeinfo : value;
      exp : int64;
      count : int;
      requested : bool;
    }


module ESet = Set.Make(struct 
                         type t = entry 
                         let compare x1 x2 = Int64.compare x1.inode x2.inode
                       end)

module I64Map = Map.Make(Int64)
  
type cache =
    { mutable by_inode : entry I64Map.t;
      mutable by_exp : ESet.t I64Map.t;
      mutable nn_client : Rpc_proxy.ManagedClient.t option;
      mutable next_tid : int;
    }

module Cache_var =
  Netplex_cenv.Make_var_type(struct type t = cache end)

let cachename = "current_inodecache"

let get_cache() =
  Cache_var.get cachename

module Rnot_proxy =
  Pfs_rpcapi_clnt.Make'Request_notifications(Rpc_proxy.ManagedClient)

module Fsys_proxy =
  Pfs_rpcapi_clnt.Make'Filesystem(Rpc_proxy.ManagedClient)

let get_nn_client cache =
  match cache.nn_client with
    | Some c -> c
    | None ->
	let conn =
	  match Netplex_cenv.lookup "Nn_manager" "RPC" with
	    | None ->
		failwith "Socket not found for Nn_manager/RPC"
	    | Some s ->
		Rpc_client.Unix s in
	let esys = (Netplex_cenv.self_cont()) # event_system in
	let mclient_config =
	  Rpc_proxy.ManagedClient.create_mclient_config
	    ~programs:[ Pfs_rpcapi_clnt.Request_notifications.V1._program;
			Pfs_rpcapi_clnt.Filesystem.V1._program;
		      ]
	    () in
	let mclient =
	  Rpc_proxy.ManagedClient.create_mclient
	    mclient_config
	    conn
	    esys in
	cache.nn_client <- Some mclient;
	mclient



let remove_entry cache e =
  cache.by_inode <- I64Map.remove e.inode cache.by_inode;
  ( try
      let el1 = I64Map.find e.exp cache.by_exp in
      let el2 = ESet.remove e el1 in
      if el2 = ESet.empty then
        cache.by_exp <- I64Map.remove e.exp cache.by_exp
      else
        cache.by_exp <- I64Map.add e.exp el2 cache.by_exp
    with Not_found -> ()
  )


let add_entry cache e =
  cache.by_inode <- I64Map.add e.inode e cache.by_inode;
  let el2 =
    try I64Map.find e.exp cache.by_exp
    with Not_found -> ESet.empty in
  cache.by_exp <- I64Map.add e.exp (ESet.add e el2) cache.by_exp


let update_inodeinfo cache esys inode =
  let c = get_nn_client cache in
  let tid = cache.next_tid in
  cache.next_tid <- tid+1;
  let tid = Int64.of_int tid in
  let e =
    Plasma_util.rpc_engine c
      Fsys_proxy.V1.begin_transaction'async tid
    ++ (fun _ ->
	  Plasma_util.rpc_engine c
	    Fsys_proxy.V1.get_inodeinfo'async (tid,inode))
    ++ (fun code ->
	  ( match code with
	      | `ok ii ->
		  (** Update the cache: *)
		  if I64Map.mem inode cache.by_inode then (
		    let e = I64Map.find inode cache.by_inode in
		    remove_entry cache e;
		    let e' = { e with inodeinfo = `Exists ii } in
		    add_entry cache e'
		  )
	      | `estale ->
		  (** Not existing: also information *)
		  if I64Map.mem inode cache.by_inode then (
		    let e = I64Map.find inode cache.by_inode in
		    remove_entry cache e;
		    let e' = { e with inodeinfo = `Deleted } in
		    add_entry cache e'
		  )
	      | #Plasma_util.errno as errno ->
		  Netlog.logf `Warning
		    "Error for get_inodeinfo(%Ld): %s" inode 
		    (Plasma_util.string_of_errno errno);
	  );
	  Plasma_util.rpc_engine c Fsys_proxy.V1.commit_transaction'async tid
       ) 
    ++ (fun _ -> eps_e (`Done()) esys)
  in
  Uq_engines.when_state
    ~is_done:(fun () -> ())
    ~is_error:(fun error ->
		 Netlog.logf `Err
		   "Exception retrieving inodeinfo: %s"
		   (Netexn.to_string error);
		 Rpc_proxy.ManagedClient.trigger_shutdown
		   c (fun () -> ());
		 ())
    e
      

let proc_inode_update sess (inode,expires) emit =
  (** Ensure this RPC can only be invoked locally: *)
  ( match Rpc_server.get_socket_name sess with
      | Unix.ADDR_UNIX _ ->
	  ()
      | _ ->
	  failwith "Cannot invoke inode_update via Internet socket"
  );

  (** For simplicity: just delete the entry *)
  Netlog.logf `Info
    "Got notification about inode update: inode=%Ld expires=%b"
    inode expires;
  let cache = get_cache() in
  ( try
      let e = I64Map.find inode cache.by_inode in
      remove_entry cache e
    with
      | Not_found ->
	  ()
  );
  emit ()


let check_is_up_to_date inode ii_check emit = 
  (** First check whether we have it: *)
  let now = Int64.of_float(Unix.time()) in
  let cache = get_cache() in
  let e =
    try
      I64Map.find inode cache.by_inode
    with Not_found ->
      { inode = inode;
	inodeinfo = `Unknown;
	exp = Int64.succ now;
	count = 0;
	requested = false;
      } in
  (** Interpret this *)
  match e.inodeinfo with
    | `Exists ii_cache ->
	emit (ii_check ii_cache)
    | `Deleted ->
	emit false
    | `Unknown ->
	(** Increase count, and if it reaches the threshold get the
	    inodeinfo from the namenode
	 *)
	let count' = e.count + 1 in
	let do_request = count' >= 3 in
	remove_entry cache e;
	let exp = if do_request then Int64.add now 10L else e.exp in
	let e' = { e with 
		     count = count';
		     requested = do_request;
		     exp = exp
		 } in
	add_entry cache e';
	if do_request then (
	  (** Request the inodeinfo *)
	  let mclient = get_nn_client cache in
	  let cont = Netplex_cenv.self_cont() in
	  let esys = cont#event_system in
	  let sockpath =
	    try List.assoc "RPC" cont#owned_container_sockets
	    with Not_found ->
	      failwith "Container socket for self not found" in
	  Rnot_proxy.V1.on_inode_update'async mclient (inode,exp,sockpath)
	    (fun get_reply ->
	       try
		 let is_ok = get_reply() in
		 (** If we can get this, also request the info itself *)
		 if is_ok then
		   update_inodeinfo cache esys inode
	       with
		 | error ->
		     Netlog.logf `Err
		       "Exception after on_inode_update: %s"
		       (Netexn.to_string error)
	    )
	);
	emit false


let proc_is_up_to_date sess (inode,ii) emit =
  check_is_up_to_date inode (fun ii_cache -> ii = ii_cache) emit


let proc_is_up_to_date_seqno sess (inode,seqno) emit =
  check_is_up_to_date inode (fun ii_cache -> seqno = ii_cache.seqno) emit


let expire cache =
  let now = Int64.of_float(Unix.time()) in
  let entries = ref ESet.empty in
  ( try
      I64Map.iter
        (fun t el ->
           if t > now then raise Exit;
           entries := ESet.union !entries el
        )
        cache.by_exp
    with
      | Exit -> ()
  );
  
  (** Really remove the entries: *)
  ESet.iter (remove_entry cache) !entries


let setup srv () =
  Pfs_rpcapi_srv.Inode_cache.V1.bind_async
    ~proc_null:(fun _ _ emit -> emit())
    ~proc_is_up_to_date
    ~proc_is_up_to_date_seqno
    srv;
  Pfs_rpcapi_srv.Notifications.V1.bind_async
    ~proc_null:(fun _ _ emit -> emit())
    ~proc_inode_update
    srv


let factory() =
  Rpc_netplex.rpc_factory
    ~configure:(fun cf cfaddr -> ())
    ~hooks:(fun () ->
	      ( object
		  inherit Netplex_kit.empty_processor_hooks()

		  method post_start_hook cont =
		    let cache = 
		      { by_inode = I64Map.empty;
			by_exp = I64Map.empty;
			nn_client = None;
			next_tid = 0;
		      } in
		    Cache_var.set cachename cache;
		    
		    let _tm =
		      Netplex_cenv.create_timer
			(fun tm -> expire cache; true)
			1.0 in
		    ()

		  method shutdown() =
		    let cache = get_cache() in
		    match cache.nn_client with
		      | Some c -> 
			  Rpc_proxy.ManagedClient.shut_down c;
			  cache.nn_client <- None
		      | None ->
			  ()
		end
	      )
	   )
    ~setup
    ~name:"nn_inodecache"
    ()
